/* -*- P4_16 -*- */
#include <core.p4>
#include <v1model.p4>

#define MAX_UINT32 0xFFFFFFFF

#define DEFAULT_RECIRC_PORT 68   // pipe local port number

#define MAX_HOPS 9
#define MAX_ENTRIES_FOR_STATE_REG 256  // 8-bit lookup key: 4 bit switch_id + 4 bit link_id 

#define MCAST_GRP_ID 1

const bit<16> ETHERTYPE_IPV4 = 0x0800;
const bit<16> ETHERTYPE_ARP	= 0x0806;
const bit<8>  PROTO_PERC = 0xFE;
const bit<8>  PROTO_PERC_RECIRC = 0xFF;
const bit<8>  PROTO_PERC_TMP = 0xFD;
const bit<8>  RESTORE_FLAG = 0xFC;

typedef bit<48> macAddr_t;
typedef bit<32> ip4Addr_t;
typedef bit<9> PortId_t;

struct port_metadata_t {
    bit<4> switch_id;
};

header ethernet_t {
    macAddr_t dstAddr;
    macAddr_t srcAddr;
    bit<16>   etherType;
}

header ipv4_t {
    bit<4>    version;
    bit<4>    ihl;
    bit<8>    diffserv;
    bit<16>   totalLen;
    bit<16>   identification;
    bit<3>    flags;
    bit<13>   fragOffset;
    bit<8>    ttl;
    bit<8>    protocol;
    bit<16>   hdrChecksum;
    ip4Addr_t srcAddr;
    ip4Addr_t dstAddr;
}

header perc_t {
    bit<32> bottleRate;  // b[l]  kbps --> 4294 Gbps
    bit<32> allocRate;   // x[l]
    bit<4>  switchID;
    bit<1>  bottleState; // s[l]
    bit<1>  ack;
    bit<1>  fin;
    bit<1>  bos;
}


header perc_recirc_t {  
	bit<32> SumE;
    bit<32> NumB;
    bit<32> link_capacity;
    bit<8>  lookup_idx;
    bit<4>  swid;
    bit<4>  is_update_pkt; // more bits to avoid PHV alloc error
    PortId_t ucast_egress_port;
    bit<7>  _pad0;
}

struct metadata_t { // 16 + 32 = 48
    port_metadata_t port_md;
    bit<4>  linkid;
    bit<32> SumE;
    bit<32> NumB;
    bit<8>  lookup_idx;
    bit<32> curr_bottleRate;
    bit<32> curr_allocRate;
    bit<32> link_capacity;
    bit<32> numerator;
    bit<32> denominator;
    bit<16> log_num;
    bit<16> log_denom;
    bit<16> exp_key;
    bit<32> b;
    bit<32> e;
    bit<32> min1;
    bit<32> min2;
    PortId_t link_port;
    bit<4>  swid;
    bit<1>  is_arp;
    bit<1>  apply_perc;
    bit<1>  curr_bottleState;
    bit<1>  need_to_update;
    bit<1>  update_already_pending;
    bit<1>  e_less_than_b;
//    bit<7>  pad0;
}


struct headers_t {
    ethernet_t  ethernet;
    ipv4_t        ipv4;
    perc_recirc_t perc_recirc;
    perc_t        perc1;
    perc_t        perc2;
    perc_t        perc3;
    perc_t        perc4;
    // perc_t[MAX_HOPS] percsTmp;
    // perc_t[MAX_HOPS] percs;
}

parser MyParser(packet_in pkt,
                out headers_t hdr,
                inout metadata_t meta,
                inout standard_metadata_t standard_metadata) {
    
    state start {
        transition parse_ethernet;
    }

    state parse_ethernet {
        pkt.extract(hdr.ethernet);
        transition select(hdr.ethernet.etherType) {
            ETHERTYPE_IPV4:      parse_ipv4;
            ETHERTYPE_ARP: parse_arp;
            default:        accept;
        }
    }

    state parse_ipv4 {
        pkt.extract(hdr.ipv4);
        transition select(hdr.ipv4.protocol) {
            PROTO_PERC:        parse_perc1;
            PROTO_PERC_RECIRC: parse_perc_recirc;
            // PROTO_PERC_TMP: parse_perc_tmp;
            // RESTORE_FLAG:   parse_perc_tmp;
            default:        accept;
        }
    }

    state parse_arp {
    	meta.is_arp = 1;
    	transition accept;
    }

    state parse_perc_recirc {
    	pkt.extract(hdr.perc_recirc);
    	transition select(hdr.perc_recirc.is_update_pkt){
    		1: accept;
    		0: parse_perc1;
    	}
    }

    state parse_perc1 {
    	pkt.extract(hdr.perc1);
    	transition select(hdr.perc1.ack){
    		0: parse_apply_perc;
    		1: accept;
    		default: accept;
    	}
    }

    state parse_apply_perc {
        meta.apply_perc = 1;
        transition select(hdr.perc1.bos){
            0: parse_perc2;
            1: accept;
            default: accept;
        }
    }

    state parse_perc2 {
    	pkt.extract(hdr.perc2);
    	transition select(hdr.perc2.bos){
    		0: parse_perc3;
    		1: accept;
    		default: accept;
    	}
    }

    state parse_perc3 {
    	pkt.extract(hdr.perc3);
    	transition select(hdr.perc3.bos){
    		0: parse_perc4;
    		1: accept;
    		default: accept;
    	}
    }

    state parse_perc4 {
    	pkt.extract(hdr.perc4);
    	transition accept;
    }

}


/*************************************************************************
************   C H E C K S U M    V E R I F I C A T I O N   *************
*************************************************************************/

control MyVerifyChecksum(inout headers_t hdr, inout metadata_t meta) {   
    apply {  }
}


/*************************************************************************
**************  I N G R E S S   P R O C E S S I N G   *******************
*************************************************************************/

control MyIngress(inout headers_t hdr,
                  inout metadata_t meta,
                  inout standard_metadata_t standard_metadata) {


    apply{}

}


/*************************************************************************
****************  E G R E S S   P R O C E S S I N G   *******************
*************************************************************************/

control MyEgress(inout headers_t hdr,
                 inout metadata_t meta,
                 inout standard_metadata_t standard_metadata) {
    apply {

    }
}

/*************************************************************************
*************   C H E C K S U M    C O M P U T A T I O N   **************
*************************************************************************/

control MyComputeChecksum(inout headers_t hdr, inout metadata_t meta) {
     apply {
    
    }
}

/*************************************************************************
***********************  D E P A R S E R  *******************************
*************************************************************************/

control MyDeparser(packet_out pkt, in headers_t hdr) {
    apply {
        pkt.emit(hdr.ethernet);
        pkt.emit(hdr.ipv4);
        pkt.emit(hdr.perc_recirc);
        pkt.emit(hdr.perc1);
        pkt.emit(hdr.perc2);
        pkt.emit(hdr.perc3);  
        pkt.emit(hdr.perc4);       
    }
}

/*************************************************************************
***********************  S W I T C H  *******************************
*************************************************************************/

V1Switch(
MyParser(),
MyVerifyChecksum(),
MyIngress(),
MyEgress(),
MyComputeChecksum(),
MyDeparser()
) main;



